import os, sys, fileinput, tempfile, gzip, tarfile, datetime
 
 
valid_data_sources = [' ', '0', '6', '7', 'B', 'C', 'K', 'U', 'X', 'Z']
valid_prc_sources  = [' ', '0', '6', '7', 'B', 'C', 'K',      'X', 'Z']

#
#  definition for an object that tells whether to use a particular
#  type of data from a station.
#
class ClassStnMeta:
    id          = ''
    name        = ''
    lat         = -9999.9
    long        = -9999.9    
    use_string  = '.......'       # tmin,tmax,prec,tavg,dewp,wind,cldc
    exclude_yrs = []

#-----------------------------------------------------
#  Parses a basic NCEI station history file that lists stations.
#  Returns a single ClassStnMeta object.
#  Since the NCEI file doesn't specify data type usage or excluded
#  years, we fill those in as "use everything".
#
def parseLineForMeta(line):
    sm = ClassStnMeta()
    try:
        sm.id   = line[0:11].strip()
        sm.name = line[41:86].strip()
        sm.lat  = float(line[11:21].strip())
        sm.long = float(line[21:31].strip())
        sm.use_string = 'YYYYYYY'
        sm.exclude_yrs = []
    except:
        sm.id   = ''
        sm.name = ''
        sm.lat  = -999.9
        sm.long = -999.9
        sm.use_string = '.......'
        sm.exclude_yrs = []
    return sm

#-----------------------------------------------------
def parseDlyLine(line):
   id   = line[0:11].strip()
   type = line[17:21].strip()
   year = int(line[11:15].strip())
   mo   = int(line[15:17].strip())
   dlyVals = [-9.9e9 for j in range(31)]
   if type=='PRCP':
       valid_src = valid_prc_sources
   else:
       valid_src = valid_data_sources

   for i in range(31):
      try:
         j = 21 + i*8
         val = int(line[j:j+5].strip())
         mFlag = line[j+5:j+6]
         qFlag = line[j+6:j+7]
         dSrc  = line[j+7:j+8]      # data source
         if dSrc in valid_src:
             if qFlag == ' ':
                dlyVals[i] = val
             else:
                dlyVals[i] = -9.9e9
      except:
         dlyVals[i] = -9.9e9
   return id, type, year, mo, dlyVals


#-----------------------------------------------------
def isLeapYear(year):
   # Is it evenly divisible by 4? If not, not a leap year
   if year%4 != 0: return False
   
   # Is it a century year? If not, it's a leap year.
   if year%100 != 0: return True
   
   # So.. it IS a century year. Then to be a leap year 
   # it must be evenly divisible by 400 (1600, 2000, 2400)
   if year%400 == 0:
      return True
   else:
      return False

#-----------------------------------------------------
def daysInMonth(mo, yr):
   if mo == 1: return 31
   if mo == 2: 
      if isLeapYear(yr):
         return 29
      else:
         return 28
   if mo == 3: return 31
   if mo == 4: return 30
   if mo == 5: return 31
   if mo == 6: return 30
   if mo == 7: return 31
   if mo == 8: return 31
   if mo == 9: return 30
   if mo == 10: return 31
   if mo == 11: return 30
   if mo == 12: return 31

#-----------------------------------------------------
def parse_exclusion_year_spec(ystr):
    try:
        s1 = ystr[:4]             # first 4 characters, which should be 1st year
        if len(ystr) > 4:
           ss = ystr.split('-')
           s2 = ss[1].strip()    # should be the end year
        else:
           s2 = s1
        sy = int(s1)
        ey = int(s2)
    except:
        sy = -1
        ey = -1
    return sy,ey

#-----------------------------------------------------
#  When using the NCEI-supplied station history file as the 
#  station list, we specify to use all data types.
def readListNCEI(filename):
    stnlist = []
    with open(filename) as listFile:
        i = 0
        for line in listFile:
           i += 1
           stnMeta = parseLineForMeta(line)
           if len(stnMeta.id) >= 11:
              stnlist.append(stnMeta)
    return stnlist

#-----------------------------------------------------
#  When using the master list, usage of data types is specified
#  in the file.
def readMasterList(filename, dtype):
    stnlist = []

    #
    #  Convert dtype string into an index integer value.
    #
    if dtype=='tmin': dnum = 1
    if dtype=='tmax': dnum = 2
    if dtype=='prec': dnum = 3
    if dtype=='tavg': dnum = 4
    if dtype=='dewp': dnum = 5
    if dtype=='wind': dnum = 6
    if dtype=='cldc': dnum = 7

    #
    #  Read the entire file into a list
    #
    with open(filename, "r", 1) as myfile:
        lines = myfile.readlines()
    
    #
    #  Find the line that marks the end of the header.
    #
    hdr_end = -1
    i = 0
    while (i < len(lines)):
        if lines[i][:10] == '----------':
            hdr_end = i
            i = len(lines)
        i = i + 1
    if hdr_end == -1:
        print('Attempt to read master list failed. Never found end of header.')
        return stnlist
        
    #
    #  Now process the lines after the header section
    #
    for line in lines[hdr_end+1:]:
        items = line.split(',')
        if len(items) >= 3:
            suo = ClassStnMeta()
            suo.id         = items[0].strip()
            suo.lat        = items[1].strip()
            suo.long       = items[2].strip()
            suo.use_string = items[3].strip()
            suo.name       = items[4].strip()
            suo.exclude_yrs = []
            if len(items) > 5:
                s = items[5].strip()
                t = s.split(';')
                for excl in t:
                    xs = excl.strip()
                    if len(xs) > 3:
                        sy,ey = parse_exclusion_year_spec(xs)
                        if sy == -1:
                           print('Unable to parse exclusion years for ' + suo.id)
                        else:
                           for y in range(sy,ey+1):
                               suo.exclude_yrs.append(y)
            ss = suo.use_string[dnum-1:dnum]      # get the single character
            if ss.lower() == 'y':
                stnlist.append(suo)

    # for debug... make sure the master list was read ok
    with open('debug_masterlist.txt', 'wt') as dbg:
        dbg.write('verification of master list read...\n')
        for suo in stnlist:
            dbg.write(suo.id.strip() + ',' + suo.use_string)
            for i in range(len(suo.exclude_yrs)):
                dbg.write(',' + str(suo.exclude_yrs[i]))
            dbg.write('\n')
    
    return stnlist

#-----------------------------------------------------
def getListOfStations(filename, dtype):
    #
    #  Read the first 10 lines to use in auto-detecting the file format
    #    
    try:
        with open(filename, "r", 1) as myfile:
            line10 = [next(myfile) for i in range(10)]
    except:
       print('Error opening the specified list file. [' + filename + ']')
       sys.exit()

    #
    #  Does the file match the NCEI station list format?
    #
    formatNCEI = True
    for i in range(10):
        line = line10[i]
        try:
            lat  = float(line[11:21].strip())
            long = float(line[21:31].strip())
            elev = float(line[31:38].strip())
        except:
            formatNCEI = False

    #
    #  If it's not in the default NCEI format, is it
    #  in our new master file format?
    #  I'm doing a real simple check here. If line 1 starts with 
    #  "Master List" I assume it's good. Case-insensitive, of course.
    #
    formatMaster = False
    if not formatNCEI:
        line = line10[0].lower()
        if line[:11] == 'master list': formatMaster = True

    #
    #  Read the list using the appropriate routine.
    #
    theList = None
    if formatNCEI:   theList = readListNCEI(filename)
    if formatMaster: theList = readMasterList(filename, dtype)
    return theList
         
#-----------------------------------------------------
#  stn is a ClassStnMeta object
#
def procDlyToMetCsv(stn):
   stnID = stn.id
   dlyFileName = stnID.strip() + '.dly'
   try:
      dlyFile = open(dlyFileName, "r", 1)
   except:
      print('Error opening the specified dly file. ['+dlyFileName+']')
      return

   #
   #  Gotta know start/end years.
   #  I will create arrays for full years of all 31-day months.
   #  That leaves some unused slots, but makes indexing easier.
   #
   syear = 9999999
   eyear = -999999
   for line in dlyFile:
      try:
         year = int(line[11:15].strip())
         if year < syear: syear = year
         if year > eyear: eyear = year
      except:
         pass
   totalDays = (eyear-syear+1)*12*31
   
   #
   #  Build a list of years that we WILL use
   #
   use_yr = []
   for yr in range(syear, eyear+1):
       if yr not in stn.exclude_yrs:
          use_yr.append(yr)
   
   #
   #  create/initialize lists
   #
   tmaxVals = [-9.9e9 for j in range(totalDays)]
   tminVals = [-9.9e9 for j in range(totalDays)]
   precVals = [-9.9e9 for j in range(totalDays)]
      
   #
   dlyFile.seek(0)    # rewind file to beginning
   i = 0
   for line in dlyFile:
      i += 1
      id, type, year, mo, dlyVals =  parseDlyLine(line)
      if year in use_yr:
          for day in range(31):            # 0..30
             k = (year-syear)*(12*31) + (mo-1)*31 + day
             if type=='TMIN': tminVals[k] = dlyVals[day] / 10.0      # tenths of C -> deg C
             if type=='TMAX': tmaxVals[k] = dlyVals[day] / 10.0      # tenths of C -> deg C
             if type=='PRCP': precVals[k] = dlyVals[day] / 10.0      # tenths of mm -> mm

   #
   #  If the stn object does not say to use the data type, set it
   #  back to missing data values.
   #
   if stn.use_string[0].lower() != 'y': 
       for i in range(len(tminVals)): tminVals[i] = -9.9e9
   if stn.use_string[1].lower() != 'y': 
       for i in range(len(tmaxVals)): tmaxVals[i] = -9.9e9
   if stn.use_string[2].lower() != 'y':
       for i in range(len(precVals)): precVals[i] = -9.9e9

   writeMetCsv(stn, syear, eyear, tmaxVals, tminVals, precVals)
   
   print('procDlyToMetCsv with ' + stnID)   

#-----------------------------------------------------
#  Write a met_*.csv file in the format required by the GLSHFS package.
#  Comma-separated entries.
#
def writeMetCsv(stn, syear, eyear, tmax, tmin, prec):
   #
   #  find the date extents for VALID (not missing) data
   #
   sdate = datetime.date(2500, 1, 1)
   edate = datetime.date(1500, 1, 1)
   for y in range(syear, eyear+1):
      for m in range(1, 13):               # 1..12
          days = daysInMonth(m,y)
          for d in range(1, days+1):       # 1..31
             k = (y-syear)*(12*31) + (m-1)*31 + d-1
             useit = False
             if tmax[k] > -998: useit = True
             if tmin[k] > -998: useit = True
             if prec[k] > -998: useit = True
             if useit:
                this_date = datetime.date(y, m, d)
                if this_date < sdate: sdate = this_date
                if this_date > edate: edate = this_date

   stnID = stn.id
   outputFilename = 'met_' + stnID.strip() + '.csv'
   try:
      f = open(outputFilename, 'w')
   except:
      print('Error opening output file ' + outputFilename)

   f.write('Daily data for a single station.\n')
   
   s = 'StationID:  , ' + stnID.strip() + '\n'
   f.write(s)

   # Station names can have imbedded commas, so enclose it in double-quotes
   s = 'StationName:, "' + stn.name.strip() + '"\n'
   f.write(s)
   
   s = 'Lat&Long    , ' + str(stn.lat) + ', ' + str(stn.long) + '\n'
   f.write(s)

   s = 'Elevation:  , -9999.000\n'
   f.write(s)

   s = 'Starts(YMD):,' + sdate.strftime("%Y-%m-%d") + '\n'
   f.write(s)

   s = 'Ends(YMD):  ,' + edate.strftime("%Y-%m-%d") + '\n'
   f.write(s)
   
   s = 'Date  ,AirTempMin, AirTempMax, Precipitation, AirTemp, Dewpoint, WindSpeed, CloudCover\n'
   f.write(s)
   
   s = 'YYYY-MM-DD, Celsius, Celsius, Millimeter, Celsius, Celsius, MetersPerSecond, Percent\n'
   f.write(s)
   
   for y in range(syear,eyear+1):
      for m in range(1,13):                   # 1..12
         days = daysInMonth(m,y)
         for d in range(1,days+1):            # 1..31
            this_date = datetime.date(y, m, d)
            if (this_date >= sdate) and (this_date <= edate):
                n = (y-syear)*31*12 + (m-1)*31 + d-1
                if tmin[n] < -999.0: 
                   v1 = "NA"
                else:
                   v1 = "%.2f" % tmin[n]
                if tmax[n] < -999.0: 
                   v2 = "NA"
                else:
                   v2 = "%.2f" % tmax[n]
                if prec[n] < -999.0: 
                   v3 = "NA"
                else:
                   v3 = "%.2f" % prec[n]
                v4 = "NA"
                v5 = "NA"
                v6 = "NA"
                v7 = "NA"
                s = "{:4d}-{:02d}-{:02d},{:s},{:s},{:s},{:s},{:s},{:s},{:s}\n".format(y,m,d,v1,v2,v3,v4,v5,v6,v7)
                f.write(s)
   f.close()

   
#-----------------------------------------------------
#  stn is a ClassStnMeta object
def procStation(stn, tarFile):
   stnID = stn.id.strip()
   try:
      stnfilename = stnID + '.dly'
      nameInTarFile = 'ghcnd_all/' + stnfilename
      member = tarFile.getmember(nameInTarFile)
      member.name = os.path.basename(member.name)
   except:
      print('Station ' + stnID + ' not found in tar file.')
      return
 

   # asList = []
   # asList.append(member)
   try:
      tarFile.extract(member, './')
   except:
      print('Error extracting station ' + stnID + ' from tar file.')
      return

   #
   #  Call external program to process the dly file?
   #  Or try doing it in python directly.
   #
   procDlyToMetCsv(stn)
   os.remove(member.name)
   
#-----------------------------------------------------
#-----------------------------------------------------
#  Main application
#
#  Command-line processing
#
numArgs = len(sys.argv)
if numArgs < 2:
   print('USAGE: process_ghcnd tarfile stnfile')
   print('   tarfile  = name of the big ghcnd tarball file downloaded from NCEI')
   print('   listfile = name of a station listing file (e.g. ghcnd-stations.txt)')
   print('Each line of the station list file will be parsed for station ID and then')
   print('that station will be extracted and processed from the tarball.')
   print('Note that the tarball can be left in compressed (gz) format, but that')
   print('will slow things down a LOT. I strongly suggest uncompressing it first')
   print('and then reading from the uncompressed tar file.')
   sys.exit()
   
tarFileName  = sys.argv[1]
listFileName = sys.argv[2]

#
#  Get the list of stations that we will attempt to extract from 
#  the tar file. The file containing this list will be in either
#  the raw station history format as it comes from NCEI or it will
#  be in the master list format that we have defined.
#  The routine will try to auto-detect the format.  If it fails,
#  though, then the procedure will stop.
#
#  stnMeta is a list of objects of type ClassStnMeta
#
stnMeta = getListOfStations(listFileName, 'prec')

#
#  Open the tar file for reading. I have this set up so it will work for
#  compressed or uncompressed tar files.
try:
   tarFile = tarfile.open(tarFileName, mode="r:")
except:
   try:
      tarFile = tarfile.open(tarFileName, mode="r:gz")
   except:      
      print('Error opening the specified tar file.')
      sys.exit()

#
#  Get a list of all the members (file names) in the tar file.
#
print('Building list of station IDs from tar file...')
try:
   memlist = tarFile.getmembers()
except:
   print('error getting member list from the tar file')
print('There are ' + str(len(memlist)) + ' entries in the list')

#
#  Process each station
#
i = 0
for stn in stnMeta:
    sfName = stn.id + '.dly'
    nameInTarFile = 'ghcnd_all/' + sfName 
    found = False
    for m in memlist:
       if m.name == nameInTarFile: 
          found = True
          break
    if found:
       procStation(stn, tarFile)
       sys.stdout.flush()
tarFile.close() 
   
