import os, sys, fileinput, tempfile, gzip, tarfile, datetime
import numpy as np
import ftplib
import csv
 
 
valid_data_sources = [' ', '0', '6', '7', 'B', 'C', 'K', 'U', 'X', 'Z']
valid_prc_sources  = [' ', '0', '6', '7', 'B', 'C', 'K',      'X', 'Z']

#
#  definition for an object that tells whether to use a particular
#  type of data from a station.
#
class ClassStnMeta:
    id          = ''
    name        = ''
    lat         = -9999.9
    long        = -9999.9    
    use_string  = '.......'       # tmin,tmax,prec,tavg,dewp,wind,cldc
    exclude_yrs = []

#-----------------------------------------------------
#  Parses a basic NCEI station history file that lists stations.
#  Returns a single ClassStnMeta object.
#  Since the NCEI file doesn't specify data type usage or excluded
#  years, we fill those in as "use everything".
#
def parseLineForMeta(line):
    sm = ClassStnMeta()
    try:
        sm.id   = line[0:11].strip()
        sm.name = line[41:86].strip()
        sm.lat  = float(line[11:21].strip())
        sm.long = float(line[21:31].strip())
        sm.use_string = 'YYYYYYY'
        sm.exclude_yrs = []
    except:
        sm.id   = ''
        sm.name = ''
        sm.lat  = -999.9
        sm.long = -999.9
        sm.use_string = '.......'
        sm.exclude_yrs = []
    return sm

#-----------------------------------------------------
def parseDlyLine(line):
   id   = line[0:11].strip()
   type = line[17:21].strip()
   year = int(line[11:15].strip())
   mo   = int(line[15:17].strip())
   dlyVals = [-9.9e9 for j in range(31)]
   if type=='PRCP':
       valid_src = valid_prc_sources
   else:
       valid_src = valid_data_sources

   for i in range(31):
      try:
         j = 21 + i*8
         val = int(line[j:j+5].strip())
         mFlag = line[j+5:j+6]
         qFlag = line[j+6:j+7]
         dSrc  = line[j+7:j+8]      # data source
         if dSrc in valid_src:
             if qFlag == ' ':
                dlyVals[i] = val
             else:
                dlyVals[i] = -9.9e9
      except:
         dlyVals[i] = -9.9e9
   return id, type, year, mo, dlyVals


#-----------------------------------------------------
def isLeapYear(year):
   # Is it evenly divisible by 4? If not, not a leap year
   if year%4 != 0: return False
   
   # Is it a century year? If not, it's a leap year.
   if year%100 != 0: return True
   
   # So.. it IS a century year. Then to be a leap year 
   # it must be evenly divisible by 400 (1600, 2000, 2400)
   if year%400 == 0:
      return True
   else:
      return False

#-----------------------------------------------------
def daysInMonth(mo, yr):
   if mo == 1: return 31
   if mo == 2: 
      if isLeapYear(yr):
         return 29
      else:
         return 28
   if mo == 3: return 31
   if mo == 4: return 30
   if mo == 5: return 31
   if mo == 6: return 30
   if mo == 7: return 31
   if mo == 8: return 31
   if mo == 9: return 30
   if mo == 10: return 31
   if mo == 11: return 30
   if mo == 12: return 31

#-----------------------------------------------------
def parse_exclusion_year_spec(ystr):
    try:
        s1 = ystr[:4]             # first 4 characters, which should be 1st year
        if len(ystr) > 4:
           ss = ystr.split('-')
           s2 = ss[1].strip()    # should be the end year
        else:
           s2 = s1
        sy = int(s1)
        ey = int(s2)
    except:
        sy = -1
        ey = -1
    return sy,ey


#-----------------------------------------------------------------
#  stnid    = single station id;  e.g. 'USC00201234'
#  stations = list of ClassStnMeta objects
#
def get_station_index(stnid, stations):
    for snum in range(len(stations)):
        stn = stations[snum]
        if stn.id == stnid: 
            return snum
    return -1            # station not found

#-----------------------------------------------------------------
# Read a master list file and return a list of ClassStnMeta objects
#
def read_stnlist(filename):
    stnlist = []

    #
    #  Read the entire file into a list
    #
    with open(filename, "r", 1) as myfile:
        lines = myfile.readlines()
    
    #
    #  Find the line that marks the end of the header.
    #
    hdr_end = -1
    i = 0
    while (i < len(lines)) and hdr_end==-1:
        if lines[i][:10] == '----------':
            hdr_end = i
        i = i + 1
    if hdr_end == -1:
        print('Attempt to read master list failed. Never found end of header.')
        print('file = [' + filename.strip() + ']')
        return stnlist
        
    #
    #  Now process the lines after the header section
    #
    for line in lines[hdr_end+1:]:
        items = line.split(',')
        if len(items) >= 3:
            suo = ClassStnMeta()
            suo.id   = items[0].strip()
            suo.lat  = items[1].strip()
            suo.long = items[2].strip()
            suo.name = items[4].strip()
            suo.use_string = items[3].strip()
            suo.exclude_yrs = []
            if len(items) > 5:
                s = items[5].strip()
                if (len(s) > 0):
                    t = s.split(';')
                    for ys in t:
                        sy,ey = parse_exclusion_year_spec(ys)
                        if sy == -1:
                           print('Unable to parse exclusion years for ' + suo.id)
                        else:
                           for y in range(sy,ey+1):
                               suo.exclude_yrs.append(y)
            stnlist.append(suo)
    return stnlist


#-----------------------------------------------------
#  Write a met_*.csv file in the format required by the GLSHFS package.
#  Comma-separated entries.
#
#  stn      = ClassStnMeta object
#  syear    = start of data
#  eyear    = end of data
#  datavals = numpy array [3, numdays]
#              [0, :] = tmin data    (degrees C)
#              [1, :] = tmax data    (degrees C)
#              [2, :] = prec data    (mm)
#
def writeMetCsv(stn, syear, eyear, datavals):
    #
    #  find the date extents for VALID (not missing) data
    #
    sdate = datetime.date(2500, 1, 1)    # ridiculous future date
    edate = datetime.date(1500, 1, 1)    # ridiculous past date
   
    datastart = datetime.date(syear, 1, 1)    # start of data in the array
    dataend   = datetime.date(eyear,12,31)    # end of data in the array
    delta = dataend - datastart
    numdays = delta.days + 1

    for d in range(numdays):
        useit = False
        if datavals[0,d] > -998.0: useit = True
        if datavals[1,d] > -998.0: useit = True
        if datavals[2,d] > -998.0: useit = True
        if useit:
            this_date = datastart + datetime.timedelta(days=d)
            if this_date < sdate: sdate = this_date
            if this_date > edate: edate = this_date

    #
    #  If there is no valid data to be written, don't write a file.
    #
    if edate < sdate: return
    
    #
    #
    #
    stnID = stn.id
    outputFilename = 'met_' + stnID.strip() + '.csv'
    print('writing ' + outputFilename)
    try:
       f = open(outputFilename, 'w')
    except:
       print('Error opening output file ' + outputFilename)

    f.write('Daily data for a single station.\n')
   
    s = 'StationID:  , ' + stnID.strip() + '\n'
    f.write(s)

    # Station names can have imbedded commas, so enclose it in double-quotes
    s = 'StationName:, "' + stn.name.strip() + '"\n'
    f.write(s)
   
    s = 'Lat&Long    , ' + str(stn.lat) + ', ' + str(stn.long) + '\n'
    f.write(s)

    s = 'Elevation:  , -9999.000\n'
    f.write(s)

    s = 'Starts(YMD):,' + sdate.strftime("%Y-%m-%d") + '\n'
    f.write(s)

    s = 'Ends(YMD):  ,' + edate.strftime("%Y-%m-%d") + '\n'
    f.write(s)
   
    s = 'Date  ,AirTempMin, AirTempMax, Precipitation, AirTemp, Dewpoint, WindSpeed, CloudCover\n'
    f.write(s)
   
    s = 'YYYY-MM-DD, Celsius, Celsius, Millimeter, Celsius, Celsius, MetersPerSecond, Percent\n'
    f.write(s)
   
    fstr = "{:4d}-{:02d}-{:02d},{:s},{:s},{:s},{:s},{:s},{:s},{:s}\n"
    numdays = (edate-sdate).days + 1
    skip = (sdate-datastart).days    # days at start of datavals with no data
    for dnum in range(numdays):
        this_date = sdate + datetime.timedelta(days=dnum)
        y = this_date.year
        m = this_date.month
        d = this_date.day
        
        n = skip + dnum                    # index of this day's data
        if datavals[0,n] < -999.0:         # tmin
            v1 = "NA"
        else:
            v1 = "{:.2f}".format(datavals[0,n])
            
        if datavals[1,n] < -999.0:         # tmax
            v2 = "NA"
        else:
            v2 = "{:.2f}".format(datavals[1,n])
            
        if datavals[2,n] < -999.0:         # prec
            v3 = "NA"
        else:
            v3 = "{:.2f}".format(datavals[2,n])
            
        v4 = "NA"
        v5 = "NA"
        v6 = "NA"
        v7 = "NA"
        
        s = fstr.format(y,m,d,v1,v2,v3,v4,v5,v6,v7)
        f.write(s)
    f.close()

#--------------------------------------------------------------------------------
def read_and_process(stnlist, datafile, syear, eyear):
    #
    #  Get the list of stations to process, including all of the metadata.
    #  This is a list of ClassStnMeta objects.
    #
    stations = read_stnlist(stnlist)

    #
    #  Create a huge numpy array that will be able to hold all of the data.
    #  If we were to read and store the data in python lists it would be 
    #  ridiculously slow, based on past experience. numpy arrays work MUCH faster.
    #
    #  This array will be indexed(station_num, dtype, daynum)
    #    station_num = 0..num_stations-1
    #    dtype       = 0..2,    where 0 = tmin
    #                                 1 = tmax
    #                                 2 = prec
    #    daynum      = 0..num_days-1
    #  The daynum index will always represent an offset from Jan 1 of year1.
    #
    num_stations = len(stations)
    sdate = datetime.date(year1,  1,  1)
    edate = datetime.date(year2, 12, 31)
    delta = edate - sdate
    num_days = delta.days + 1
    
    print('num_stations=', num_stations)
    print('num_days    =', num_days)
    
    all_data = np.full((num_stations, 3, num_days), -9.9e29)

    #
    #  Process the file, storing each data value into the correct spot in
    #  the all_data array.
    #
    dtype_list = ['tmin', 'tmax', 'prcp']
    with open(datafile) as csv_file:
        csv_reader = csv.reader(csv_file, delimiter=',')
        line_count = 0
        for row in csv_reader:
            line_count += 1
            if line_count % 500000 == 0:
                print('Processing line ', line_count)
            try:
                stnid     = row[0]
                date_str  = row[1]
                dtype_str = row[2].lower()
                value     = float(row[3])
                snum  = get_station_index(stnid, stations)
                if snum >= 0:
                    try:
                        dtype = dtype_list.index(dtype_str)    # data type index              
                    except:
                        dtype = -1
                    if dtype >= 0:
                        y = int(date_str[0:4])
                        m = int(date_str[4:6])
                        d = int(date_str[6:8])
                        dt = datetime.date(y,m,d)
                        delta = dt - sdate
                        day = delta.days       # 0..num_days-1
                    
                        #
                        #  The data values in the file are:
                        #    Tmin & Tmax : tenths of degrees C
                        #    Precip      : tenths of mm
                        #  Convert both from the tenths to just plain C or mm
                        #
                        all_data[snum,dtype,day] = value/10.0
            except:
                print('error splitting line ', line_count)
                print('row=', row)
                print(' ')
                pass
    print('Processed ', line_count, ' lines')
    
    #
    #  For each station, write the output file
    #
    for snum in range(num_stations):
        stn = stations[snum]
        stn_data = all_data[snum,:,:]
        writeMetCsv(stn, syear, eyear, stn_data)
    
   
#-----------------------------------------------------
#-----------------------------------------------------
#  Main application
#
#  Command-line processing
#
#  Main application
#
#  Command-line processing
#
numArgs = len(sys.argv)
if numArgs < 4:
   print('USAGE: python process_ghcnd_nrt.py filename year1 year2')
   print('   filename = name of a master-list format file')
   print('   year1 = earliest year to retrieve/process')
   print('   year2 = latest year to retrieve/process')
   print('')
   print('The specified year(s) will be used to control the download of ')
   print('files from the NCEI ftp site. The GHCND data is stored there in')
   print('a "by_year" directory such that each yearly file contains ALL data')
   print('for a particular year.')
   print('The master list will then be used to control the processing of stations')
   print('from those downloaded files. The result will be a single met_*.csv file')
   print('for each station.')
   sys.exit()
   
listfile = sys.argv[1]
year1 = int(sys.argv[2])
year2 = int(sys.argv[3])

ok = True
if year1 < 1763:  ok = False
if year2 > 2029:  ok = False
if year2 < year1: ok = False
if not ok:
   print('Invalid year(s) specified')
   print('The GHCND site set has data from 1763-present')
   sys.exit()

#
#  Download all of the yearly files.
#  Concatenate them all together into a single giant file.
#  In many cases this would be a bad thing for efficiency, but in
#  this case it really doesn't matter. Every line in every file is
#  a single observation for a single station and we are gonna have to 
#  parse the entire dataset.  This simplifies that part of the process.
#
datafile = 'lumped_ghcnd_near_real_time.txt'
lumpF = open(datafile, 'wb')
for y in range (year1, year2+1):
    yrStr = str(y)
    gzName = yrStr.strip() + '.csv.gz'    # e.g. 1995.csv.gz
    print('downloading ' + gzName)
    ftp = ftplib.FTP('ftp.ncdc.noaa.gov', 'anonymous', 'tim.hunter@noaa.gov')
    ftp.cwd('/pub/data/ghcn/daily/by_year/')
    worked = False
    try:
        ftp.retrbinary('RETR %s' % gzName, open("temp.gz", "wb").write)
        worked = True
    except:
        worked = False
    ftp.quit()
    if worked:
        with gzip.open("temp.gz", "rb") as z:
            file_lines = z.readlines()
            for line in file_lines:
                lumpF.write(line)
        os.remove("temp.gz")
lumpF.close()

#
#  read and process the data
#
read_and_process(listfile, datafile, year1, year2)

